from .core import AverMarket, AverClient, AverOutcome
from solana.rpc.api import Client
from solana.publickey import PublicKey
from pyserum.market import Market as SerumMarket
from pyserum.enums import OrderType, Side
from pyserum.open_orders_account import OpenOrdersAccount

import base64
import base58
from pyserum.connection import conn

from solana.rpc.types import RPCResponse, TxOpts
from solana.account import Account
from solana.transaction import Transaction
from spl.token.client import Token
from spl.token._layouts import ACCOUNT_LAYOUT
from spl.token.instructions import get_associated_token_address, create_associated_token_account
import math
import datetime as dt
import pandas as pd



class AverDeXMarket():

    def __init__(
        self,
        aver_market: AverMarket,
        auto_mint: bool = False,
        auto_settle: bool = False,
        ):
        pass
        self.aver_market = aver_market
        self.dex_outcomes = [AverDeXOutcome(o) for o in self.aver_market.outcomes]
    
    # Functions that require looking across all outcome tokens...


class AverDeXOutcome(AverOutcome):
    '''
        AverDeXOutcome Class
        A subclass of the AverOutcome class, which can be initiated where
          an Aver market's outcomes are tradeable on AverDeX/Serum.
        Facilitates many of the commonly required interactions with
          AverDeX/Serum, layering in additional userful functionality 
          specific to the nature of Aver betting markets.
    '''

    def __init__(self, aver_outcome: AverOutcome):
        self.aver_outcome = aver_outcome
        self.serum_market = SerumMarket.load(
            conn=self.aver_outcome.aver_market.client.connection,
            market_address=self.aver_outcome.serum_market_address,
            program_id=self.aver_outcome.serum_program_id
        )   
        self.quote_token_mint = self.aver_outcome.aver_market.base_token_mint
        self.outcome_ata = self.aver_outcome.user_ata
        self.quote_ata = self.aver_outcome.aver_market.user_base_token_ata
        self.outcome_lot_size = self.serum_market.state.base_lot_size()
        self.quote_lot_size = self.serum_market.state.quote_lot_size()
        self.outcome_spl_token_multiplier = self.serum_market.state.base_spl_token_multiplier()
        self.quote_spl_token_multiplier = self.serum_market.state.quote_spl_token_multiplier()
        self.increment_size = self.outcome_lot_size/self.outcome_spl_token_multiplier
        self.price_tick_size = self.quote_lot_size/self.quote_spl_token_multiplier
        self.quote_dp = -int(round(math.log(self.price_tick_size, 10),0))
        self.outcome_dp = -int(round(math.log(self.increment_size, 10),0))
        self.open_orders_accounts = self.serum_market.find_open_orders_accounts_for_owner(PublicKey(self.aver_outcome.aver_market.client.user_account.public_key()))
        self.open_orders_addresses = [str(o.address) for o in self.open_orders_accounts]

    def price_lots_to_number(self, price: int) -> float:
        return float((price * self.quote_lot_size * self.outcome_spl_token_multiplier) / (self.outcome_lot_size * self.quote_spl_token_multiplier))

    def price_number_to_lots(self, price: float) -> int:
        return int(round((price * self.quote_spl_token_multiplier * self.outcome_lot_size) / (self.outcome_spl_token_multiplier() * self.quote_lot_size)))

    def outcome_size_lots_to_number(self, size: int) -> float:
        return float(size * self.outcome_lot_size) / self.outcome_spl_token_multiplier

    def outcome_size_number_to_lots(self, size: float) -> int:
        return int(math.floor(size * self.outcome_spl_token_multiplier) / self.outcome_lot_size)

    def quote_size_lots_to_number(self, size: int) -> float:
        return float(size * self.quote_lot_size) / self.quote_spl_token_multiplier

    def quote_size_number_to_lots(self, size: float) -> int:
        return int(math.floor(size * self.quote_spl_token_multiplier) / self.quote_lot_size)

    def load_bids(self):
        self.bids = self.serum_market.load_bids()
        return self.bids
    
    def load_asks(self):
        self.asks = self.serum_market.load_asks()
        return self.asks

    def get_outcome_wallet_balance(self):
        self.outcome_wallet_balance = self.aver_outcome.balance()
        return self.outcome_wallet_balance 

    def get_quote_wallet_balance(self):
        self.quote_wallet_balance = self.aver_outcome.aver_market.base_token_balance()
        return self.quote_wallet_balance

    def get_outcome_open_orders_free_balance(self):
        self.outcome_open_orders_free_balance = sum([o.base_token_free for o in self.open_orders_accounts])/self.outcome_spl_token_multiplier
        return self.outcome_open_orders_free_balance

    def get_quote_open_orders_free_balance(self):
        self.quote_open_orders_free_balance = sum([o.quote_token_free for o in self.open_orders_accounts])/self.quote_spl_token_multiplier
        return self.quote_open_orders_free_balance 
        
    def get_outcome_open_orders_total_balance(self):
        print([o.base_token_total for o in self.open_orders_accounts])
        self.outcome_open_orders_total_balance = sum([o.base_token_total for o in self.open_orders_accounts])/self.outcome_spl_token_multiplier
        return self.outcome_open_orders_total_balance

    def get_quote_open_orders_total_balance(self):
        self.quote_open_orders_total_balance = sum([o.quote_token_total for o in self.open_orders_accounts])/self.quote_spl_token_multiplier
        return self.quote_open_orders_total_balance

    def get_open_orders(self):
        self.open_orders = self.serum_market.load_orders_for_owner(PublicKey(self.aver_outcome.aver_market.client.user_account.public_key()))
        return self.open_orders    

    def cancel_all_open_orders(self):
        open_orders = self.get_open_orders()
        cancellations = []
        for order in open_orders:
            cancellation = self.serum_market.cancel_order(
                owner=self.aver_market.client.user_account,
                order=order,
            )
            cancellations += [cancellation]
        return cancellations

    def cancel_order(self, order):
        cancellation = self.serum_market.cancel_order(
            owner=self.aver_outcome.aver_market.client.user_account,
            order=order,
        )
        return cancellation

    def place_order(
        self,
        side: Side,
        price: float,
        quantity: float,
        order_type: OrderType = OrderType.LIMIT,
        client_id: int = 0,
        check: bool = True,
        mint: bool = None,
        ):
        '''
            Places an order to BACK (buy) or LAY (sell)
            If check=True, balance checks will be performed before executing the order to
              to ensure sufficient tokens and SOL to make complete the transaction
            If mint=True, or auto_mint=True at market level, then the function will request
              a minting of the necessary outcome tokensets to fulfil the LAY order before
              executing the order on the DeX.
        '''

        #if mint == None:
        #    mint = self.aver_market.auto_mint

        if side == Side.BUY:
            payer = self.quote_ata
            wallet_balance = self.get_quote_wallet_balance()
            required_tokens = price * quantity
        else:
            payer = self.outcome_ata
            wallet_balance = self.get_outcome_wallet_balance()
            required_tokens = quantity

        if check:
            if self.aver_outcome.aver_market.client.get_lamports_balance() < self.aver_outcome.aver_market.client.fees_per_signature:
                print(f'ERROR: Did not trade - insufficient lamports | Balance: {self.lamports_wallet_balance}')
                return None
            elif wallet_balance < required_tokens:
                if side == Side.SELL and mint:
                    additional_tokens_required = int((required_tokens - wallet_balance) *(10**(self.aver_outcome.aver_market.decimals)))/(10**(self.aver_outcome.aver_market.decimals))
                    if self.aver_market.get_base_token_balance() >= additional_tokens_required*(1+self.aver_outcome.aver_market.mint_fee):
                        print(f'Auto-minting outcome tokensets to meet Lay order | Sets required: {required_tokens} Balance: {wallet_balance} | Additional {additional_tokens_required}')
                        self.aver_outcome.aver_market.mint(additional_tokens_required)
                    else:
                        print(f'ERROR: Did not trade - insufficient tokens, tried auto-minting | Required: {required_tokens} Balance: {wallet_balance}')
                        return None
                else:
                    print(f'ERROR: Did not trade - insufficient tokens | Required: {required_tokens} Balance: {wallet_balance}')
                    return None

        order_txn = self.serum_market.place_order( 
            payer=payer,
            owner=self.aver_outcome.aver_market.client.user_account,
            order_type=order_type, 
            side=side,  
            limit_price=price,
            max_quantity=quantity,
            client_id=client_id,
            opts=TxOpts(skip_confirmation=self.aver_outcome.aver_market.client.skip_confirmation),
        )

        return order_txn


    def settle_balances(self):
        '''
            Settle balances for completed trades
        '''
        
        for open_order_account in self.open_orders_accounts:
            
            settlement_txn = self.serum_market.settle_funds( 
                owner=self.aver_outcome.aver_market.client.user_account,
                open_orders=open_order_account,
                base_wallet=self.outcome_ata,
                quote_wallet=self.quote_ata,  
            )
            return settlement_txn

    
    def crank_market(self):
        '''
            Cranks the market to ensure that event queue is processed,
              orders are matched, and balances can be settled.
            Not required on mainnet, but necessary for devnet and testnet
        '''
        events = self.serum_market.load_event_queue()
        open_orders_intial = list(map(lambda e : e.public_key, events))
        open_orders = list(reversed(open_orders_intial))
        limit = 10
        if len(open_orders) >= 3:
            crank_txn = self.serum_market._conn.send_transaction(
                self.serum_market._build_consume_events_tx(open_orders, limit),
                self.aver_outcome.aver_market.client.user_account,
                opts=TxOpts(skip_confirmation=self.aver_outcome.aver_market.client.skip_confirmation)
            )
            return crank_txn
        else:
            return None