from .instructions import *
from .utils import await_confirmation
from typing import List, NamedTuple, Optional
from solana.rpc.api import Client
import solana.rpc.types as types
from solana.system_program import transfer, TransferParams, create_account, CreateAccountParams
from solana.account import Account 
from solana.publickey import PublicKey 
from solana.transaction import Transaction, AccountMeta, TransactionInstruction
from solana.instruction import decode_data, encode_data, InstructionLayout
from spl.token.instructions import get_associated_token_address, create_associated_token_account
from spl.token._layouts import MINT_LAYOUT, ACCOUNT_LAYOUT
from spl.token.instructions import (
    mint_to, MintToParams, initialize_account, InitializeAccountParams,
)
import struct
import requests
import datetime as dt
import json
import base64
import base58



AVER_BASE_URL = 'https://aver.bet/'


class AverClient():
    '''
        Aver Client Class
        An Aver Client is unique to a Solana network and user combination   
        If no user account is provided, a generic Account() will be created:
        - It can be used to obtain information from Aver or the AverDeX.
        - However, this will be unable to trade as it has no SOL or token balances.    
    '''

    def __init__(
        self,
        solana_client: str,
        solana_account: Account = Account(),
        skip_confirmation: bool = False
        ) -> None:

        self.connection = Client(solana_client)
        self.user_account = solana_account
        self.user_public_key = PublicKey(self.user_account.public_key())
        self.skip_confirmation = skip_confirmation
        self.base_url = AVER_BASE_URL
        self.fees_per_signature = self.connection.get_fees()['result']['value']['feeCalculator']['lamportsPerSignature']
        self.rent_for_new_ata = float(self.connection.get_minimum_balance_for_rent_exemption(ACCOUNT_LAYOUT.sizeof())['result'])
        
    #def list_active_events(self):
    #    return requests.get(AVER_BASE_URL+'/api/events', data = {'status':'active'}).json()['result']
    
    def list_active_minting_markets(self):
        return requests.get(AVER_BASE_URL+'/api/markets', data = {'status':'active_minting'}).json()['result']

    def list_active_trading_markets(self):
        return requests.get(AVER_BASE_URL+'/api/markets', data = {'status':'active_trading'}).json()['result']
 
    def list_resolved_markets(self):
        return requests.get(AVER_BASE_URL+'/api/markets', data = {'status':'resolved'}).json()['result']

    #def get_event(self, event_id: int):
    #    return requests.get(AVER_BASE_URL+f'/api/event/{event_id}').json()['result']

    def list_event_markets(self, event_id: int):
        return requests.get(AVER_BASE_URL+f'/api/event/{event_id}/markets').json()['result']
    
    def get_market(self, market_id: int):
        data = requests.get(AVER_BASE_URL+f'/api/market/{market_id}').json()['result']
        return data

    def get_outcome(self, outcome_id: int):
        return requests.get(AVER_BASE_URL+f'/api/outcome/{outcome_id}').json()['result']

    def list_base_tokens(self):
        return requests.get(AVER_BASE_URL+f'/api/base_tokens').json()['result']

    #def get_base_tokens(self, base_token_name: str):
    #    return requests.get(AVER_BASE_URL+f'/api/base_token/{base_token_name}').json()['result']

    def request_base_token_airdrop(self, base_token_name: str, amount: float):
        '''Only works on DEVNET and TESTNET'''
        return requests.post(AVER_BASE_URL+f'/api/airdrop/', data={'currency':base_token_name, 'pubkey':self.user_account.public_key(), 'amount':amount}).json()

    def request_lamports_airdrop(self, amount: float):
        '''Only works on DEVNET and TESTNET'''
        resp = {}
        while 'result' not in resp:
            resp = self.connection.request_airdrop(self.connection, amount*10**9)
        await_confirmation(self.connection, resp['result'])

    def get_lamports_balance(self):
        account_info = self.connection.get_account_info(self.user_public_key)
        try:
            return float(account_info['result']['value']['lamports'])
        except:
            return 0

    def get_token_balance(self, token_mint: PublicKey):
        ata = get_associated_token_address(self.user_public_key, token_mint)
        try:
            balance = float(self.connection.get_token_account_balance(ata)['result']['value']['uiAmount'])
        except:
            balance = 0
        return balance

    def get_base_token_balance(self, base_token_name: str):
        base_token_data = self.get_base_tokens(base_token_name)
        return self.get_token_balance(base_token_data['token_mint'])

    def get_outcome_token_balance(self, outcome_id: int):
        outcome_data = self.get_outcome(outcome_id)
        return self.get_token_balance(PublicKey(outcome_data['token_mint'])) 

    #def get_base_tokens_held(self, user_pubkey: PublicKey):
    #   Query Aver to check any balances in the user's wallet of Aver recognized based tokens
    #    return...

    #def get_outcome_tokens_held(self, user_pubkey: PublicKey):
    #   Query Aver to check any balances in the user's wallet of Aver recognized outcome tokens
    #    return...

    #def get_active_outcome_tokens_held(self, user_pubkey: PublicKey):
    #   Query Aver to check any balances in the user's wallet of Aver recognized ooutcome tokens for active markets
    #    return...

    #def get_resolved_outcome_tokens_held(self, user_pubkey: PublicKey):
    #   Query Aver to check any balances in the user's wallet of Aver recognized ooutcome tokens for resolved markets
    #    return...



class AverMarket():
    '''
        Aver Market Class
        Corresponds one-to-one with an Aver market contract on the blockchain.
        The information necessary to interact with the market's instance, as well as context data are
          retrieved from the Aver API to initiate the class instance.

    '''
    def __init__(
        self,
        aver_client: AverClient, 
        aver_market_id: int,
        ):

        self.client = aver_client
        self.market_id = aver_market_id
        self.get_market_info()

    def get_market_info(self):
        '''
            Sends a request to obtain market context and Aver protocol 
              transaction parameters from the Aver API 
        '''
        self.data = self.client.get_market(self.market_id)
        print(self.data)
        self.name = self.data['name']
        self.description = self.data['description']
        self.bet_type = self.data['bet_type']
        self.network = self.data['network']
        self.base_token_name = self.data['base_token_name']
        self.base_token_mint = PublicKey(self.data['base_token_mint'])
        self.user_base_token_ata = get_associated_token_address(PublicKey(self.client.user_account.public_key()), self.base_token_mint)
        self.decimals = self.data['decimals']
        self.fee_mint = self.data['fee_mint'] 
        self.fee_redeem = self.data['fee_redeem'] 
        self.fee_collect = self.data['fee_collect'] 
        self.resolution_start = dt.datetime.strptime(self.data['resolution_start'].replace('T', ' '), "%Y-%m-%d %H:%M:%S%z")
        self.resolution_end = dt.datetime.strptime(self.data['resolution_end'].replace('T', ' '), "%Y-%m-%d %H:%M:%S%z")
        self.active_minting = self.data['active_minting']
        self.active_minting = self.data['active_trading']
        self.resolved = self.data['resolved']
        self.aver_program_id = self.data['aver_program_id']
        self.aver_market_account = PublicKey(self.data['aver_market_account'])
        self.aver_escrow_account = PublicKey(self.data['aver_escrow_account'])
        self.aver_fee_account = PublicKey(self.data['aver_fee_account'])
        self.aver_escrow_authority_account = PublicKey(self.data['aver_escrow_authority_account'])
        self.token_program_id = PublicKey(self.data['token_program_id'])

        self.outcomes = []
        for o in self.data['outcomes']:
            self.outcomes +=  [AverOutcome(
                aver_market = self,
                id = o['id'],
                name = o['name'],
                token_mint = PublicKey(o['token_mint']),
                serum_program_id = PublicKey(o['serum_program_id']),
                serum_market_address = PublicKey(o['serum_market']),
                user_ata = get_associated_token_address(PublicKey(self.client.user_account.public_key()), PublicKey(o['token_mint']))
            )]
        self.number_outcomes = len(self.outcomes)
        if self.resolved:
            self.winning_outcome_id = self.data['winning_outcome_id']
            self.winning_outcome_name = self.data['winning_outcome_name'] 
            self.winning_outcome_mint = PublicKey(self.data['winning_outcome_mint'])


    def base_token_balance(self):
        return self.client.get_token_balance(self.base_token_mint)

    def create_ata_instructions(self, pubkeys=None) -> Transaction:
        '''
            Creates instructions to initiate ATAs for outcome token
            or base token mints not yet initiated for the user's primary account.
        '''
        if pubkeys == None:    # If none provided specifically, try for all ATAs required in this market
            pubkeys = [self.base_token_mint] + [o.token_mint for o in self.outcomes] 

        tx = Transaction()
        for mint_account in pubkeys:
            ata = get_associated_token_address(PublicKey(self.client.user_account.public_key()), PublicKey(mint_account))
            ata_info = self.client.connection.get_account_info(ata)['result']['value']
            if ata_info is not None: 
                account_state = ACCOUNT_LAYOUT.parse(base64.b64decode(ata_info['data'][0])).state
            else:
                account_state = 0
            if account_state == 0:
                ata_ix = create_associated_token_account(
                    payer=PublicKey(self.client.user_account.public_key()), 
                    owner=PublicKey(self.client.user_account.public_key()),
                    mint=PublicKey(mint_account),
                )
                tx = tx.add(ata_ix)
        return tx
        
        
    def create_atas_if_required(self):
        '''
            Sends transaction to network to create ATAs for all outcome token mints and
              base token mint which have not yet been initiated.
            No transaction sent if all ATAs already exist.
        '''
        tx = self.create_ata_instructions()

        if len(tx.instructions) > 0: # i.e. if required, send transactions
            signers = [self.client.user_account]
            try:
                response = self.client.connection.send_transaction(tx, *signers, opts=types.TxOpts(skip_confirmation=self.client.skip_confirmation))
                return response['result']['transaction']['signatures']
            except Exception as e:
                print(e)
                raise(e)
        else:
            return None
    

    def request_base_token_airdrop(self, amount):
        '''
            Sends a reques to the Aver API asking for an airdrop of the market's base token.
            Note: Does not work on MAINNET.
        '''
        self.client.request_base_token_airdrop(self.base_token_name, amount)


    def mint(self, amount):
        '''
            Creates request to mint a specified amount of outcome tokensets for the market
            If associated token accounts are required, creation of these are included in the transaction
            Checks to see if sufficient base tokens and lamports are available to cover the transaction before proceeding
        '''

        # Check the user has enought base tokens to mint
        base_token_balance = self.client.get_token_balance(self.base_token_mint)
        if base_token_balance < amount * (1+self.fee_mint):
            print(f'Insufficient base tokens to mint this number of outcome tokensets ({self.name} -> {self.base_token_name})')
            return None

        tx = Transaction()
        lamport_cost = 0

        # Check the user has a base token ATA, otherwise include an instruction to create one
        ata_ix = self.create_ata_instructions()
        if len(ata_ix.instructions) != 0: 
            tx = tx.add(ata_ix)
            lamport_cost += (self.client.fees_per_signature + self.client.rent_for_new_ata) * len(ata_ix.instructions)
    
        tx = tx.add(
            mint_tokens(
                MintTokensParams(
                    program_id= self.aver_program_id,
                    aver_market_account= self.aver_market_account,
                    escrow_account= self.aver_escrow_account,
                    fee_account= self.aver_fee_account,
                    outcome_token_mints= [o.token_mint for o in self.outcomes],
                    user_account= PublicKey(self.client.user_account.public_key()),
                    user_base_token_account= self.user_base_token_ata,
                    user_outcome_token_accounts= [o.user_ata for o in self.outcomes],
                    escrow_authority_account= self.aver_escrow_authority_account,
                    token_program_id= self.token_program_id,
                    amount= int(amount),
                    decimals= int(self.decimals),
                )
            )
        )
            
        signers = [self.client.user_account]

        lamport_cost += self.client.fees_per_signature * len(signers)
        lamport_balance = self.client.get_lamports_balance()

        if lamport_balance < lamport_cost:
            print(f'Insufficient lamports (Balance: {lamport_balance} | Required: {lamport_cost}) to carry out this transaction')
            return None

        try:
            response = self.client.connection.send_transaction(tx, *signers, opts=types.TxOpts(skip_confirmation=self.client.skip_confirmation))
            return response['result']['transaction']['signatures']
        except Exception as e:
            raise(e)


    def redeem(self, amount: float):
        '''
            Creates request to redeem a specified amount of outcome tokensets for the market
            If associated token accounts are required, creation of these are included in the transaction
            Checks to see if sufficient complete outcome tokensets and lamports are available to cover the transaction before proceeding
        '''

        # Check the user has enought complete tokensets to redeem
        outcome_token_balances = [self.client.get_token_balance(o.token_mint) for o in self.outcomes]
        if min(outcome_token_balances) < amount:
            print(f'Insufficient complete tokensets to mint this number of outcome tokensets ({self.name} -> {self.base_token_name})')
            return None

        tx = Transaction()
        lamport_cost = 0

        # Check the user has a base token ATA, otherwise include an instruction to create one
        ata_ix = self.create_ata_instructions()
        if len(ata_ix.instructions) != 0: 
            tx = tx.add(ata_ix)
            lamport_cost += (self.client.fees_per_signature + self.client.rent_for_new_ata) * len(ata_ix.instructions)
    

        tx = tx.add(
            redeem_tokens(
                RedeemTokensParams(
                    program_id= self.aver_program_id,
                    aver_market_account= self.aver_market_account,
                    escrow_account= self.aver_escrow_account,
                    fee_account= self.aver_fee_account,
                    outcome_token_mints= [o.token_mint for o in self.outcomes],
                    user_account= PublicKey(self.client.user_account.public_key()),
                    user_base_token_account= self.user_base_token_ata,
                    user_outcome_token_accounts= [o.user_ata for o in self.outcomes],
                    escrow_authority_account= self.aver_escrow_authority_account,
                    token_program_id= self.token_program_id,
                    amount= int(amount),
                    decimals= int(self.decimals),
                )
            )
        )
            
        signers = [self.client.user_account]

        lamport_cost += self.client.fees_per_signature * len(signers)
        lamport_balance = self.client.get_lamports_balance()

        if lamport_balance < lamport_cost:
            print(f'Insufficient lamports (Balance: {lamport_balance} | Required: {lamport_cost}) to carry out this transaction')
            return None

        try:
            response = self.client.connection.send_transaction(tx, *signers, opts=types.TxOpts(skip_confirmation=self.client.skip_confirmation))
            return response['result']['transaction']['signatures']
        except Exception as e:
            raise(e)
    

    def collect(self):
        '''
            Creates request to collect any winning tokens held by the user for this market
            If an associated token accounts is required (for the base token), creation of this ATA is included in the transaction
            Checks to see if the market has been resolved; if any winning tokens are help; and if sufficient lamports are available to cover the transaction before proceeding
        '''

        if not self.resolved:
            print(f'The market ({self.name}) has not yet been resolved.')
            return None

        else:
            # Check the user has winning tokens to collect
            winning_token_balance = self.client.get_token_balance(self.winning_outcome_mint)
            if winning_token_balance == 0:
                print(f'No winning outcome tokens ({self.name} -> {self.winning_outcome_name}) to collect.')
                return None
            
            tx = Transaction()
            lamport_cost = 0

            # Check the user has a base token ATA, otherwise include an instruction to create one
            base_ata_ix = self.create_ata_instructions([self.base_token_mint])
            if len(base_ata_ix.instructions) != 0: 
                tx = tx.add(base_ata_ix)
                lamport_cost += (self.client.fees_per_signature + self.client.rent_for_new_ata) * len(base_ata_ix.instructions)
            
            tx = tx.add(
                redeem_tokens(
                    RedeemTokensParams(
                        program_id= self.aver_program_id,
                        aver_market_account= self.aver_market_account,
                        escrow_account= self.aver_escrow_account,
                        fee_account= self.aver_fee_account,
                        outcome_token_mints= [o.token_mint for o in self.outcomes],
                        user_account= PublicKey(self.client.user_account.public_key()),
                        user_base_token_account= self.user_base_token_ata,
                        user_outcome_token_accounts= [o.user_ata for o in self.outcomes],
                        escrow_authority_account= self.aver_escrow_authority_account,
                        token_program_id= self.token_program_id,
                        decimals= int(self.decimals),
                    )
                )
            )

            signers = [self.client.user_account]

            lamport_cost += self.client.fees_per_signature * len(signers)
            lamport_balance = self.client.get_lamports_balance()

            if lamport_balance < lamport_cost:
                print(f'Insufficient lamports (Balance: {lamport_balance} | Required: {lamport_cost}) to carry out this transaction')
                return None

            try:
                response = self.client.connection.send_transaction(tx, *signers, opts=types.TxOpts(skip_confirmation=self.client.skip_confirmation))
                return response['result']['transaction']['signatures']
            except Exception as e:
                raise(e)
            
        
class AverOutcome():
    '''
        Aver Outcome Class
        A chile of the Aver Market Class - corresponds one-to-one with
          the outcome token mints of a given Aver Market.
    '''
    def __init__(
        self,
        aver_market: AverMarket,
        id: int,
        name: str,
        token_mint: PublicKey,
        serum_program_id: PublicKey,
        serum_market_address: PublicKey,
        user_ata: PublicKey = None,
        ):
        self.aver_market = aver_market
        self.id = id
        self.name = name
        self.token_mint = token_mint
        self.serum_program_id = serum_program_id
        self.serum_market_address = serum_market_address
        self.user_ata = user_ata

    def balance(self):
        '''
            Returns the balance of this outcome token held by the user
        '''
        return self.aver_market.client.get_token_balance(self.token_mint)

    def collect(self):
        '''
            Requests a collection if the market is resolved and this outcome token is the
              market's winning outcome token
        '''
        if self.aver_market.resolved:
            if self.aver_market.winning_outcome_mint == self.token_mint:
                self.aver_market.collect()
            else:
                print(f'This outcome token ({self.name}) was not the winning token of this market ({self.aver_market.name}).')
        else:
            print(f'This market ({self.aver_market.name}) has not yet been resolved.')
