from typing import List, NamedTuple
from solana.publickey import PublicKey
from solana.transaction import AccountMeta, TransactionInstruction
import struct

class MintTokensParams(NamedTuple):
    """Mint Aver Token Params"""

    program_id: PublicKey

    aver_market_account: PublicKey

    escrow_account: PublicKey

    fee_account: PublicKey

    outcome_token_mints: List[PublicKey]

    user_account: PublicKey

    user_base_token_account: PublicKey

    user_outcome_token_accounts: List[PublicKey]

    escrow_authority_account: PublicKey

    token_program_id: PublicKey

    amount: int

    decimals: int


class RedeemTokensParams(NamedTuple):
    """Redeem Aver Contract Params"""

    program_id: PublicKey

    aver_market_account: PublicKey

    escrow_account: PublicKey

    fee_account: PublicKey

    outcome_token_mints: List[PublicKey]

    user_account: PublicKey

    user_base_token_account: PublicKey

    user_outcome_token_accounts: List[PublicKey]

    escrow_authority_account: PublicKey

    token_program_id: PublicKey

    amount: int

    decimals: int


class CollectTokensParams(NamedTuple):
    """Collect Aver Contract Params"""

    program_id: PublicKey

    aver_market_account: PublicKey

    escrow_account: PublicKey

    fee_account: PublicKey

    outcome_token_mints: List[PublicKey]

    user_account: PublicKey

    user_base_token_account: PublicKey

    user_outcome_token_accounts: List[PublicKey]

    escrow_authority_account: PublicKey

    token_program_id: PublicKey



def mint_tokens(params: MintTokensParams) -> TransactionInstruction:
    """Generate a transaction instruction to mint tokens on an Aver market."""

    keys = [
            AccountMeta(pubkey=params.aver_market_account, is_signer=False, is_writable=True),
            AccountMeta(pubkey=params.escrow_account, is_signer=False, is_writable=True),
            AccountMeta(pubkey=params.fee_account, is_signer=False, is_writable=True),
        ]
    keys += [AccountMeta(pubkey=x, is_signer=False, is_writable=True) for x in params.outcome_token_mints]
    keys += [
        AccountMeta(pubkey=params.user_account, is_signer=True, is_writable=False),
        AccountMeta(pubkey=params.user_base_token_account, is_signer=False, is_writable=True),
    ]
    keys += [AccountMeta(pubkey=x, is_signer=False, is_writable=True) for x in params.user_outcome_token_accounts]
    keys += [
        AccountMeta(pubkey=params.escrow_authority_account, is_signer=False, is_writable=False),
        AccountMeta(pubkey=params.token_program_id, is_signer=False, is_writable=False),
    ]
    
    data = struct.pack("<BQ", 1, int(params.amount*(10**params.decimals)))

    return TransactionInstruction(
        keys=keys,
        program_id=params.program_id,
        data=data,
    )        


def redeem_tokens(params: RedeemTokensParams) -> TransactionInstruction:
    """Generate a transaction instruction to redeem tokens on an Aver market."""

    keys = [
            AccountMeta(pubkey=params.aver_market_account, is_signer=False, is_writable=True),
            AccountMeta(pubkey=params.user_account, is_signer=False, is_writable=False),
    ]
    keys += [AccountMeta(pubkey=x, is_signer=False, is_writable=True) for x in params.user_outcome_token_accounts]
    keys += [AccountMeta(pubkey=params.user_base_token_account, is_signer=False, is_writable=True)]
    keys += [AccountMeta(pubkey=x, is_signer=False, is_writable=True) for x in params.outcome_token_mints]
    keys += [
        AccountMeta(pubkey=params.escrow_account, is_signer=False, is_writable=True),
        AccountMeta(pubkey=params.fee_account, is_signer=False, is_writable=True),
        AccountMeta(pubkey=params.escrow_authority_account, is_signer=False, is_writable=True),
        AccountMeta(pubkey=params.token_program_id, is_signer=False, is_writable=False),
    ]

    data = struct.pack("<BQ", 2, int(params.amount*(10**params.decimals)))

    return TransactionInstruction(
        keys=keys,
        program_id=params.program_id,
        data=data,
    )


def collect_tokens(params: CollectTokensParams) -> TransactionInstruction:
    """Generate a transaction instruction to collect tokens on an Aver market."""

    keys = [
            AccountMeta(pubkey=params.aver_market_account, is_signer=False, is_writable=True),
            AccountMeta(pubkey=params.user_account, is_signer=True, is_writable=False),
    ]
    keys += [AccountMeta(pubkey=x, is_signer=False, is_writable=True) for x in params.user_outcome_token_accounts]
    keys += [AccountMeta(pubkey=params.user_base_token_account, is_signer=False, is_writable=True)]
    keys += [AccountMeta(pubkey=x, is_signer=False, is_writable=True) for x in params.outcome_token_mints]
    keys += [
        AccountMeta(pubkey=params.escrow_account, is_signer=False, is_writable=True),
        AccountMeta(pubkey=params.fee_account, is_signer=False, is_writable=True),
        AccountMeta(pubkey=params.escrow_authority_account, is_signer=False, is_writable=False),
        AccountMeta(pubkey=params.token_program_id, is_signer=False, is_writable=False),
    ]

    data = struct.pack("<B", 4)

    return TransactionInstruction(
        keys=keys,
        program_id=params.program_id,
        data=data,
    )
